@@ -44,6 +44,7 @@
 #include <QtMultimedia/qvideosurfaceformat.h>
 #include <QtMultimedia/qcameraimagecapture.h>
 #include <private/qmemoryvideobuffer_p.h>
+#include <private/qvideoframe_p.h>
 
 #include "dscamerasession.h"
 #include "dsvideorenderer.h"
@@ -67,6 +68,7 @@ DSCameraSession::DSCameraSession(QObject *parent)
     , m_previewStarted(false)
     , m_surface(nullptr)
     , m_previewPixelFormat(QVideoFrame::Format_Invalid)
+    , m_stride(-1)
     , m_readyForCapture(false)
     , m_imageIdCounter(0)
     , m_currentImageId(-1)
@@ -130,30 +132,28 @@ void DSCameraSession::setViewfinderSettings(const QCameraViewfinderSettings &set
 qreal DSCameraSession::scaledImageProcessingParameterValue(
         const ImageProcessingParameterInfo &sourceValueInfo)
 {
-    if (sourceValueInfo.currentValue == sourceValueInfo.defaultValue) {
+    if (sourceValueInfo.currentValue == sourceValueInfo.defaultValue)
         return 0.0f;
-    } else if (sourceValueInfo.currentValue < sourceValueInfo.defaultValue) {
+    if (sourceValueInfo.currentValue < sourceValueInfo.defaultValue) {
         return ((sourceValueInfo.currentValue - sourceValueInfo.minimumValue)
                 / qreal(sourceValueInfo.defaultValue - sourceValueInfo.minimumValue))
                 + (-1.0f);
-    } else {
-        return ((sourceValueInfo.currentValue - sourceValueInfo.defaultValue)
-                / qreal(sourceValueInfo.maximumValue - sourceValueInfo.defaultValue));
     }
+    return ((sourceValueInfo.currentValue - sourceValueInfo.defaultValue)
+            / qreal(sourceValueInfo.maximumValue - sourceValueInfo.defaultValue));
 }
 
 qint32 DSCameraSession::sourceImageProcessingParameterValue(
         qreal scaledValue, const ImageProcessingParameterInfo &valueRange)
 {
-    if (qFuzzyIsNull(scaledValue)) {
+    if (qFuzzyIsNull(scaledValue))
         return valueRange.defaultValue;
-    } else if (scaledValue < 0.0f) {
+    if (scaledValue < 0.0f) {
         return ((scaledValue - (-1.0f)) * (valueRange.defaultValue - valueRange.minimumValue))
                 + valueRange.minimumValue;
-    } else {
-        return (scaledValue * (valueRange.maximumValue - valueRange.defaultValue))
-                + valueRange.defaultValue;
     }
+    return (scaledValue * (valueRange.maximumValue - valueRange.defaultValue))
+            + valueRange.defaultValue;
 }
 
 static QCameraImageProcessingControl::ProcessingParameter searchRelatedResultingParameter(
@@ -439,9 +439,6 @@ bool DSCameraSession::unload()
 
     setStatus(QCamera::UnloadingStatus);
 
-    m_previewSampleGrabber->deleteLater();
-    m_previewSampleGrabber = nullptr;
-
     m_needsHorizontalMirroring = false;
     m_supportedViewfinderSettings.clear();
     m_supportedFormats.clear();
@@ -482,7 +479,7 @@ bool DSCameraSession::startPreview()
     if (m_surface)
         m_surface->start(m_previewSurfaceFormat);
 
-    hr = m_filterGraph->QueryInterface(IID_IMediaControl, (void**)&pControl);
+    hr = m_filterGraph->QueryInterface(IID_IMediaControl, reinterpret_cast<void**>(&pControl));
     if (FAILED(hr)) {
         errorString = tr("Failed to get stream control");
         goto failed;
@@ -504,7 +501,7 @@ failed:
     if (m_surface && m_surface->isActive())
         m_surface->stop();
     disconnectGraph();
-    setError(QCamera::CameraError, errorString);
+    setError(QCamera::CameraError, errorString, hr);
     return false;
 }
 
@@ -520,7 +517,8 @@ bool DSCameraSession::stopPreview()
 
     QString errorString;
     IMediaControl* pControl = 0;
-    HRESULT hr = m_filterGraph->QueryInterface(IID_IMediaControl, (void**)&pControl);
+    HRESULT hr = m_filterGraph->QueryInterface(IID_IMediaControl,
+                                               reinterpret_cast<void**>(&pControl));
     if (FAILED(hr)) {
         errorString = tr("Failed to get stream control");
         goto failed;
@@ -542,12 +540,13 @@ bool DSCameraSession::stopPreview()
     return true;
 
 failed:
-    setError(QCamera::CameraError, errorString);
+    setError(QCamera::CameraError, errorString, hr);
     return false;
 }
 
-void DSCameraSession::setError(int error, const QString &errorString)
+void DSCameraSession::setError(int error, const QString &errorString, HRESULT hr)
 {
+    qErrnoWarning(hr, "[0x%x] %s", hr, qPrintable(errorString));
     emit cameraError(error, errorString);
     setStatus(QCamera::UnloadedStatus);
 }
@@ -585,10 +584,13 @@ int DSCameraSession::captureImage(const QString &fileName)
         return m_imageIdCounter;
     }
 
+    const QString ext = !m_imageEncoderSettings.codec().isEmpty()
+        ? m_imageEncoderSettings.codec().toLower()
+        : QLatin1String("jpg");
     m_imageCaptureFileName = m_fileNameGenerator.generateFileName(fileName,
                                                          QMediaStorageLocation::Pictures,
                                                          QLatin1String("IMG_"),
-                                                         QLatin1String("jpg"));
+                                                         ext);
 
     updateReadyForCapture();
 
@@ -606,16 +608,9 @@ void DSCameraSession::onFrameAvailable(double time, const QByteArray &data)
 
     m_presentMutex.lock();
 
-    // If no frames provided from ISampleGrabber for some time
-    // the device might be potentially unplugged.
-    m_deviceLostEventTimer.start(100);
-
-    // (We should be getting only RGB32 data)
-    int stride = m_previewSize.width() * 4;
-
     // In case the source produces frames faster than we can display them,
     // only keep the most recent one
-    m_currentFrame = QVideoFrame(new QMemoryVideoBuffer(data, stride),
+    m_currentFrame = QVideoFrame(new QMemoryVideoBuffer(data, m_stride),
                                  m_previewSize,
                                  m_previewPixelFormat);
 
@@ -639,6 +634,10 @@ void DSCameraSession::onFrameAvailable(double time, const QByteArray &data)
 
 void DSCameraSession::presentFrame()
 {
+    // If no frames provided from ISampleGrabber for some time
+    // the device might be potentially unplugged.
+    m_deviceLostEventTimer.start(100);
+
     m_presentMutex.lock();
 
     if (m_currentFrame.isValid() && m_surface) {
@@ -654,17 +653,11 @@ void DSCameraSession::presentFrame()
     m_captureMutex.lock();
 
     if (m_capturedFrame.isValid()) {
-        Q_ASSERT(m_previewPixelFormat == QVideoFrame::Format_RGB32);
 
-        m_capturedFrame.map(QAbstractVideoBuffer::ReadOnly);
+        captureImage = qt_imageFromVideoFrame(m_capturedFrame);
 
-        captureImage = QImage(m_capturedFrame.bits(),
-                                m_previewSize.width(), m_previewSize.height(),
-                                QImage::Format_RGB32);
-
-        captureImage = captureImage.mirrored(m_needsHorizontalMirroring); // also causes a deep copy of the data
-
-        m_capturedFrame.unmap();
+        const bool needsVerticalMirroring = m_previewSurfaceFormat.scanLineDirection() != QVideoSurfaceFormat::TopToBottom;
+        captureImage = captureImage.mirrored(m_needsHorizontalMirroring, needsVerticalMirroring); // also causes a deep copy of the data
 
         QtConcurrent::run(this, &DSCameraSession::processCapturedImage,
                           m_currentImageId, m_captureDestinations, captureImage, m_imageCaptureFileName);
@@ -688,8 +681,9 @@ void DSCameraSession::processCapturedImage(int id,
                                            const QImage &image,
                                            const QString &path)
 {
+    const QString format = m_imageEncoderSettings.codec();
     if (captureDestinations & QCameraImageCapture::CaptureToFile) {
-        if (image.save(path, "JPG")) {
+        if (image.save(path, !format.isEmpty() ? format.toUtf8().constData() : "JPG")) {
             Q_EMIT imageSaved(id, path);
         } else {
             Q_EMIT captureError(id, QCameraImageCapture::ResourceError,
@@ -714,7 +708,7 @@ bool DSCameraSession::createFilterGraph()
 
     // Create the filter graph
     hr = CoCreateInstance(CLSID_FilterGraph,NULL,CLSCTX_INPROC,
-            IID_IGraphBuilder, (void**)&m_filterGraph);
+                          IID_IGraphBuilder, reinterpret_cast<void**>(&m_filterGraph));
     if (FAILED(hr)) {
         errorString = tr("Failed to create filter graph");
         goto failed;
@@ -722,7 +716,8 @@ bool DSCameraSession::createFilterGraph()
 
     // Create the capture graph builder
     hr = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC,
-                          IID_ICaptureGraphBuilder2, (void**)&m_graphBuilder);
+                          IID_ICaptureGraphBuilder2,
+                          reinterpret_cast<void**>(&m_graphBuilder));
     if (FAILED(hr)) {
         errorString = tr("Failed to create graph builder");
         goto failed;
@@ -756,7 +751,8 @@ bool DSCameraSession::createFilterGraph()
                     QString output = QString::fromWCharArray(strName);
                     mallocInterface->Free(strName);
                     if (m_sourceDeviceName.contains(output)) {
-                        hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**)&m_sourceFilter);
+                        hr = pMoniker->BindToObject(nullptr, nullptr, IID_IBaseFilter,
+                                                    reinterpret_cast<void**>(&m_sourceFilter));
                         if (SUCCEEDED(hr)) {
                             pMoniker->Release();
                             break;
@@ -775,7 +771,8 @@ bool DSCameraSession::createFilterGraph()
                     while (pEnum->Next(1, &pMoniker, NULL) == S_OK) {
                         IPropertyBag *pPropBag = 0;
 
-                        hr = pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void**)(&pPropBag));
+                        hr = pMoniker->BindToStorage(nullptr, nullptr, IID_IPropertyBag,
+                                                     reinterpret_cast<void**>(&pPropBag));
                         if (FAILED(hr)) {
                             pMoniker->Release();
                             continue; // Don't panic yet
@@ -783,7 +780,8 @@ bool DSCameraSession::createFilterGraph()
 
                         // No need to get the description, just grab it
 
-                        hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**)&m_sourceFilter);
+                        hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter,
+                                                    reinterpret_cast<void**>(&m_sourceFilter));
                         pPropBag->Release();
                         pMoniker->Release();
                         if (SUCCEEDED(hr)) {
@@ -807,9 +805,9 @@ bool DSCameraSession::createFilterGraph()
 
     // Sample grabber filter
     if (!m_previewSampleGrabber) {
-        m_previewSampleGrabber = new DirectShowSampleGrabber;
+        m_previewSampleGrabber = new DirectShowSampleGrabber(this);
         connect(m_previewSampleGrabber, &DirectShowSampleGrabber::bufferAvailable,
-                this, &DSCameraSession::onFrameAvailable);
+                this, &DSCameraSession::onFrameAvailable, Qt::DirectConnection);
     }
 
 
@@ -832,7 +830,7 @@ failed:
     SAFE_RELEASE(m_nullRendererFilter);
     SAFE_RELEASE(m_filterGraph);
     SAFE_RELEASE(m_graphBuilder);
-    setError(QCamera::CameraError, errorString);
+    setError(QCamera::CameraError, errorString, hr);
 
     return false;
 }
@@ -841,9 +839,11 @@ bool DSCameraSession::configurePreviewFormat()
 {
     // Resolve viewfinder settings
     int settingsIndex = 0;
+    const QSize captureResolution = m_imageEncoderSettings.resolution();
+    const QSize resolution = captureResolution.isValid() ? captureResolution : m_viewfinderSettings.resolution();
     QCameraViewfinderSettings resolvedViewfinderSettings;
     for (const QCameraViewfinderSettings &s : qAsConst(m_supportedViewfinderSettings)) {
-        if ((m_viewfinderSettings.resolution().isEmpty() || m_viewfinderSettings.resolution() == s.resolution())
+        if ((resolution.isEmpty() || resolution == s.resolution())
                 && (qFuzzyIsNull(m_viewfinderSettings.minimumFrameRate()) || qFuzzyCompare((float)m_viewfinderSettings.minimumFrameRate(), (float)s.minimumFrameRate()))
                 && (qFuzzyIsNull(m_viewfinderSettings.maximumFrameRate()) || qFuzzyCompare((float)m_viewfinderSettings.maximumFrameRate(), (float)s.maximumFrameRate()))
                 && (m_viewfinderSettings.pixelFormat() == QVideoFrame::Format_Invalid || m_viewfinderSettings.pixelFormat() == s.pixelFormat())
@@ -868,29 +868,40 @@ bool DSCameraSession::configurePreviewFormat()
     VIDEOINFOHEADER *videoInfo = reinterpret_cast<VIDEOINFOHEADER*>(m_sourceFormat->pbFormat);
     videoInfo->AvgTimePerFrame = 10000000 / resolvedViewfinderSettings.maximumFrameRate();
 
-    // We only support RGB32, if the capture source doesn't support
-    // that format, the graph builder will automatically insert a
-    // converter.
+    m_previewPixelFormat = resolvedViewfinderSettings.pixelFormat();
+    const AM_MEDIA_TYPE *resolvedGrabberFormat = &m_sourceFormat;
 
-    if (m_surface && !m_surface->supportedPixelFormats(QAbstractVideoBuffer::NoHandle)
-            .contains(QVideoFrame::Format_RGB32)) {
-        qWarning() << "Video surface needs to support RGB32 pixel format";
-        return false;
+    if (m_surface) {
+        const auto surfaceFormats = m_surface->supportedPixelFormats(QAbstractVideoBuffer::NoHandle);
+        if (!surfaceFormats.contains(m_previewPixelFormat)) {
+            if (surfaceFormats.contains(QVideoFrame::Format_RGB32)) {
+                // As a fallback, we support RGB32, if the capture source doesn't support
+                // that format, the graph builder will automatically insert a
+                // converter (when possible).
+
+                static const AM_MEDIA_TYPE rgb32GrabberFormat { MEDIATYPE_Video, MEDIASUBTYPE_ARGB32, 0, 0, 0, FORMAT_VideoInfo, nullptr, 0, nullptr};
+                resolvedGrabberFormat = &rgb32GrabberFormat;
+                m_previewPixelFormat = QVideoFrame::Format_RGB32;
+
+            } else {
+                qWarning() << "Video surface needs to support at least RGB32 pixel format";
+                return false;
+            }
+        }
     }
 
-    m_previewPixelFormat = QVideoFrame::Format_RGB32;
     m_previewSize = resolvedViewfinderSettings.resolution();
     m_previewSurfaceFormat = QVideoSurfaceFormat(m_previewSize,
                                                  m_previewPixelFormat,
                                                  QAbstractVideoBuffer::NoHandle);
-    m_previewSurfaceFormat.setScanLineDirection(QVideoSurfaceFormat::BottomToTop);
+    m_previewSurfaceFormat.setScanLineDirection(DirectShowMediaType::scanLineDirection(m_previewPixelFormat, videoInfo->bmiHeader));
+    m_stride = DirectShowMediaType::bytesPerLine(m_previewSurfaceFormat);
 
     HRESULT hr;
     IAMStreamConfig* pConfig = 0;
-    hr = m_graphBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
-                                       &MEDIATYPE_Video,
-                                       m_sourceFilter,
-                                       IID_IAMStreamConfig, (void**)&pConfig);
+    hr = m_graphBuilder->FindInterface(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video,
+                                       m_sourceFilter, IID_IAMStreamConfig,
+                                       reinterpret_cast<void**>(&pConfig));
     if (FAILED(hr)) {
         qWarning() << "Failed to get config for capture device";
         return false;
@@ -905,9 +916,7 @@ bool DSCameraSession::configurePreviewFormat()
         return false;
     }
 
-    // Set sample grabber format
-    static const AM_MEDIA_TYPE grabberFormat { MEDIATYPE_Video, MEDIASUBTYPE_ARGB32, 0, 0, 0, FORMAT_VideoInfo, nullptr, 0, nullptr};
-    if (!m_previewSampleGrabber->setMediaType(&grabberFormat))
+    if (!m_previewSampleGrabber->setMediaType(resolvedGrabberFormat))
         return false;
 
     m_previewSampleGrabber->start(DirectShowSampleGrabber::CallbackMethod::BufferCB);
@@ -1064,8 +1073,8 @@ void DSCameraSession::updateSourceCapabilities()
 
     IAMVideoControl *pVideoControl = 0;
     hr = m_graphBuilder->FindInterface(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video,
-                                       m_sourceFilter,
-                                       IID_IAMVideoControl, (void**)&pVideoControl);
+                                       m_sourceFilter, IID_IAMVideoControl,
+                                       reinterpret_cast<void**>(&pVideoControl));
     if (FAILED(hr)) {
         qWarning() << "Failed to get the video control";
     } else {
@@ -1091,8 +1100,8 @@ void DSCameraSession::updateSourceCapabilities()
     }
 
     hr = m_graphBuilder->FindInterface(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video,
-                                       m_sourceFilter,
-                                       IID_IAMStreamConfig, (void**)&pConfig);
+                                       m_sourceFilter, IID_IAMStreamConfig,
+                                       reinterpret_cast<void**>(&pConfig));
     if (FAILED(hr)) {
         qWarning() << "failed to get config on capture device";
         return;
@@ -1169,4 +1178,23 @@ void DSCameraSession::updateSourceCapabilities()
     updateImageProcessingParametersInfos();
 }
 
+QList<QSize> DSCameraSession::supportedResolutions(bool *continuous) const
+{
+    if (continuous)
+        *continuous = false;
+
+    QList<QSize> res;
+    for (auto &settings : m_supportedViewfinderSettings) {
+        auto size = settings.resolution();
+        if (!res.contains(size))
+            res << size;
+    }
+
+    std::sort(res.begin(), res.end(), [](const QSize &r1, const QSize &r2) {
+        return qlonglong(r1.width()) * r1.height() < qlonglong(r2.width()) * r2.height();
+    });
+
+    return res;
+}
+
 QT_END_NAMESPACE
